import random
import enum
from collections import namedtuple
from math import sin
from typing import TYPE_CHECKING, Tuple, Any

import numpy
import pygame
from pygame.surface import Surface


if TYPE_CHECKING:
    from robingame.objects.entity import Entity

Point = namedtuple("Point", ["x", "y"])


def touching(entity1: "Entity", entity2: "Entity") -> bool:
    """Determine whether entity1 is touching entity2. This function is directional by
    design---i.e. it checks whether
        (a) entity1.touchbox collides with entity2.rect
        it does *not* check whether
        (b) entity2.touchbox collides with entity1.rect.

    This is because in most cases the entities will have touchboxes with the same width
    buffer. If this is the case then (a) cannot be true without (b) also being true,
    and we needn't bother doing both calculations.

    Also, I want to allow for the possibility of unequal touchbox buffer widths; in
    this case, the order *will* matter. The entity with the larger touchbox buffer will
    "touch" the other entity, but not vice versa."""
    return entity1.touchbox.colliderect(entity2.rect)


def get_overlap_between_objects(obj1, obj2):
    """Get the x and y overlap between obj1 and obj2.rect"""
    x_overlap = min(obj1.rect.right, obj2.rect.right) - max(obj1.rect.left, obj2.rect.left)
    y_overlap = min(obj1.rect.bottom, obj2.rect.bottom) - max(obj1.rect.top, obj2.rect.top)
    return x_overlap, y_overlap


def un_overlap(movable_object, immovable_object):
    """Move movable_object outside boundaries of immovable_object."""
    x_overlap, y_overlap = get_overlap_between_objects(movable_object, immovable_object)
    if x_overlap > y_overlap:
        if movable_object.centroid.y >= immovable_object.centroid.y:
            movable_object.rect.top = immovable_object.rect.bottom
        else:
            movable_object.rect.bottom = immovable_object.rect.top
    else:
        if movable_object.centroid.x >= immovable_object.centroid.x:
            movable_object.rect.left = immovable_object.rect.right
        else:
            movable_object.rect.right = immovable_object.rect.left


# need these helper functions because pygame.mask.Mask methods aren't actually
# implemented as they are described in the docs...
def maskFromSurface(surface, threshold=127):
    # return pygame.mask.from_surface(surface, threshold)

    mask = pygame.mask.Mask(surface.get_size())
    key = surface.get_colorkey()
    if key:
        for y in range(surface.get_height()):
            for x in range(surface.get_width()):
                if surface.get_at((x, y)) != key:
                    mask.set_at((x, y), 1)
    else:
        for y in range(surface.get_height()):
            for x in range(surface.get_width()):
                if surface.get_at((x, y))[3] > threshold:
                    mask.set_at((x, y), 1)
    return mask


def mask_to_surface(mask, set_color=None):
    set_color = set_color if set_color else pygame.color.THECOLORS["magenta"]
    width, height = mask.get_size()
    surface = pygame.Surface((width, height)).convert_alpha()
    surface.fill((0, 0, 0, 0))
    for y in range(height):
        for x in range(width):
            if mask.get_at((x, y)):
                surface.set_at((x, y), set_color)
    return surface


def outline_image(surface: Surface):
    mask = maskFromSurface(surface)
    return mask.outline()


def ticks_to_frames(tick, ticks_per_frame):
    return tick // ticks_per_frame


def draw_arrow(surface, origin, angle_deg, color=None, length=50):
    """angle needs to be in degrees"""
    arrow_xy = arrow_coords(length, length // 10, length // 4)
    r = rotation_matrix(angle_deg)
    arrow_xy = arrow_xy.dot(r)
    arrow_xy += numpy.array(origin)
    pygame.draw.polygon(surface, color, arrow_xy, 2)


def draw_rect(surface, color, rect, width=0):
    # make a surface with exactly the same dimensions as the screen
    surface_with_alpha = pygame.Surface((surface.get_width(), surface.get_height())).convert_alpha()
    surface_with_alpha.fill((0, 0, 0, 0))
    # pygame.draw *does* respect alpha values when you're not plotting to screen
    pygame.draw.rect(surface_with_alpha, color, rect, width)
    # now blit the whole surface to the screen
    surface.blit(surface_with_alpha, (0, 0))

    # todo: now generalise this so that it can do pygame's other draw utils. Can i make a
    #  decorator for pygame's draw functions? Or a wrapper around the whole pygame.draw module?


def rotation_matrix(angle_deg):
    angle_rad = numpy.deg2rad(angle_deg)
    rotation_matrix = numpy.array(
        [
            (numpy.cos(angle_rad), -numpy.sin(angle_rad)),
            (numpy.sin(angle_rad), numpy.cos(angle_rad)),
        ]
    )
    return rotation_matrix


def arrow_coords(length, width, head_length):
    return numpy.array(
        [
            (0, 0),
            (length, 0),
            (length - head_length, width),
            (length - head_length, -width),
            (length, 0),
        ]
    )


def count_edges(values):
    rising_edges = 0
    falling_edges = 0
    for ii, value in enumerate(values):
        if ii == 0:
            previous_value = value
        if value and not previous_value:
            rising_edges += 1
        if not value and previous_value:
            falling_edges += 1
        previous_value = value
    return rising_edges, falling_edges


def pulsing_value(tick, min, max, freq, func=None):
    func = func if func else sin
    A = (max - min) / 2
    B = (max + min) / 2
    return A * func(tick * freq) + B


def mouse_hovering_over(element):
    mouse_x, mouse_y = pygame.mouse.get_pos()
    return element.rect.collidepoint(mouse_x, mouse_y)


class MyEnumMeta(enum.EnumMeta):
    """
    Incredibly, enum.Enum does not have a __contains__ method, so we can't do stuff like:
    class Colours(Enum):
        ...
    assert "red" in Colours
    """

    def __contains__(cls, member):
        return any(x.value == member for x in cls)


class Enum(enum.Enum, metaclass=MyEnumMeta):
    pass


def circle_surf(radius, color):
    radius = int(radius)
    surf = Surface((radius * 2, radius * 2))
    pygame.draw.circle(surf, color, (radius, radius), radius)
    surf.set_colorkey((0, 0, 0))
    return surf


def random_float(min, max):
    assert min < max
    spread = max - min
    value = min + random.random() * spread
    return value


def random_int(min, max):
    return random.randint(min, max)


def limit_value(value, between: Tuple[Any, Any]):
    _min, _max = between
    return max(_min, min(value, _max))


def unzip(zipped):
    return tuple(zip(*zipped))


Coord = tuple[int, int]


class SparseMatrix(dict):
    """
    Dictionary of the form {(x, y): value}. Can be used to simulate an infinite grid.
    Each entry represents a cell with width and height of 1 (i.e. NOT a dimensionless point)
    """

    Limit = tuple[int, int] | tuple[None, None]

    @property
    def limits(self) -> tuple[Limit, Limit]:
        """The min and max x/y coords of all entries"""
        if self:
            xs, ys = unzip(self.keys())
            xlim = min(xs), max(xs)
            ylim = min(ys), max(ys)
        else:
            xlim = ylim = None, None
        return xlim, ylim

    @property
    def size(self) -> (int, int):
        if self:
            xlim, ylim = self.limits
            width = xlim[1] - xlim[0] + 1
            height = ylim[1] - ylim[0] + 1
        else:
            width = height = 0
        return width, height

    @property
    def xlim(self) -> Limit:
        xlim, _ = self.limits
        return xlim

    @property
    def ylim(self) -> Limit:
        _, ylim = self.limits
        return ylim

    @property
    def width(self) -> int:
        width, _ = self.size
        return width

    @property
    def height(self) -> int:
        _, height = self.size
        return height

    def copy(self) -> "SparseMatrix":
        return SparseMatrix(super().copy())

    def scale_to_screen(self, screen_size: (int, int)) -> (float, int, int):
        """
        Calculate the scaling factor and x/y offset required to blit a group of pixels
        representing self onto a screen
        """
        screen_width, screen_height = screen_size
        self_width, self_height = self.size
        xlim, ylim = self.limits
        x_scaling = screen_width // self_width
        y_scaling = screen_height // self_height
        scaling = max(1, min(x_scaling, y_scaling))

        x_offset = screen_width // 2 - (xlim[0] + self_width // 2) * scaling
        y_offset = screen_height // 2 - (ylim[0] + self_height // 2) * scaling
        return scaling, x_offset, y_offset

    def screen_coords(self, xy, scaling, x_offset, y_offset) -> tuple[int, int]:
        """
        Convert col/row number (xy) into screen coordinates (in pixels)
        """
        x, y = xy
        screen_x = x_offset + x * scaling
        screen_y = y_offset + y * scaling
        return screen_x, screen_y


def draw_text(s: str, surface: Surface, position, font, color, antialias=False):
    text_bitmap = font.render(s, antialias, color)
    surface.blit(text_bitmap, position)



from dataclasses import dataclass

import pygame.mouse

from robingame.gui.button import Button
from robingame.input import EventQueue
from robingame.objects import Game, Group
from robingame.utils import mouse_hovering_over


class OnPressEventExample(Game):
    """
    This is an example where the hooks increment/decrement a value stored on the outer Game class.
    """

    screen_color = (50, 50, 50)

    def __init__(self):
        super().__init__()
        self.value = 0
        self.buttons = Group()
        self.particles = Group()
        self.child_groups += [self.buttons, self.particles]
        self.buttons.add(
            Button(
                x=200,
                y=100,
                width=200,
                height=50,
                text="change value",
                on_press=lambda button: EventQueue.add(ChangeValue(amount=5)),
                on_release=lambda button: EventQueue.add(ChangeValue(amount=-5)),
                on_focus=lambda button: EventQueue.add(ChangeValue(amount=1)),
                on_unfocus=lambda button: EventQueue.add(ChangeValue(amount=-1)),
            )
        )

    def update(self):
        super().update()

        # listen for events
        if event := EventQueue.get(type=ChangeValue.type):
            self.value += event.amount
        print(f"{self.value=}")

        # button manager stuff
        for button in self.buttons:
            if mouse_hovering_over(button):
                button.is_focused = True
                button.is_pressed = pygame.mouse.get_pressed()[0]
            else:
                button.is_focused = False
                button.is_pressed = False


@dataclass
class ChangeValue:
    type = pygame.event.custom_type()
    amount: int = 1


if __name__ == "__main__":
    OnPressEventExample().main()



from pathlib import Path

from pygame.color import Color
from pygame.surface import Surface

from robingame.image import load_spritesheet, scale_image, empty_image
from robingame.text.exceptions import TextError


class Font:
    letters: dict[str:Surface]
    image_size: tuple[int, int]
    xpad: int
    ypad: int

    def __init__(
        self,
        filename: str | Path,
        image_size: tuple[int, int],
        letters: str,
        xpad=0,
        ypad=0,
        trim=False,
        space_width=None,
        **kwargs,
    ):
        """`letters` is the letters in the spritesheet, in the same order."""
        self.image_size = width, height = image_size
        self.xpad = xpad
        self.ypad = ypad
        self.letters = dict()
        self.letters[" "] = empty_image((space_width or width, height))
        self.not_found = Surface(image_size)
        self.not_found.fill(Color("red"))
        images = load_spritesheet(filename, image_size=image_size, **kwargs)
        if trim:
            images = self.trim_images(images)
        self.letters.update({letter: image for letter, image in zip(letters, images)})

    def render(
        self,
        surf: Surface,
        text: str,
        x: int = 0,
        y: int = 0,
        scale: int = 1,
        wrap: int = 0,
        align: int = None,
    ) -> int:
        """
        align: -1=left, 0=center, 1=right
        wrap: x width at which to wrap text
        """
        _, ysize = self.image_size
        cursor = x
        for line in text.splitlines():
            wrapped_lines = self.wrap_words(line, wrap, x, scale) if wrap else [line]
            for line in wrapped_lines:
                cursor = self.align_cursor(line, x, align, scale, wrap)
                for letter in line:
                    image = self.get(letter)
                    image = scale_image(image, scale)
                    surf.blit(image, (cursor, y))
                    w = image.get_width()
                    cursor += w + self.xpad * scale
                y += (ysize + self.ypad) * scale
        return cursor

    def align_cursor(self, line: str, x: int, align: int, scale: int, wrap: int) -> int:
        """Used for left/right/centered text alignmnent"""
        match align:
            case -1 | None:
                cursor = x
            case 0:
                if not wrap:
                    raise TextError("Can't center text without specifying a wrap width.")
                line_width = self.printed_width(line, scale)
                slack = wrap - line_width
                cursor = x + slack // 2
            case 1:
                line_width = self.printed_width(line, scale)
                cursor = x + wrap - line_width
            case _:
                raise TextError(f"Bad alignment value: {align}")
        return cursor

    def wrap_words(self, text: str, wrap: int, x: int = 0, scale: int = 1) -> list[str]:
        """Break one long line into multiple lines based on the wrap width."""
        lines = []
        line = ""
        for word in text.split(" "):
            new_line = f"{line} {word}" if line else word
            if self.printed_width(new_line, scale) <= wrap:
                line = new_line
            else:
                lines.append(line)
                line = word
        lines.append(line)  # last line
        return lines

    def printed_width(self, text: str, scale: int) -> int:
        return sum((self.get(letter).get_width() + self.xpad) * scale for letter in text)

    def trim_images(self, images: list[Surface]) -> list[Surface]:
        """Make a monospaced font non-monospaced"""
        trimmed = []
        for image in images:
            x, _, w, _ = image.get_bounding_rect()  # trim x to bounding rect
            _, y, _, h = image.get_rect()  # maintain original y position of character
            new = image.subsurface((x, y, w, h))
            trimmed.append(new)
        return trimmed

    def get(self, letter: str) -> Surface:
        try:
            return self.letters[letter]
        except KeyError:
            return self.not_found



from typing import Callable

import pygame
from pygame import Color
from pygame.rect import Rect

from robingame.objects import PhysicalEntity
from robingame.utils import pulsing_value


class Button(PhysicalEntity):
    """
    GUI button entity. Reacts to changes in the is_focused and is_pressed booleans, but doesn't
    set them (a parent menu class or something else should set them). This is so that each Button
    doesn't have to do its own input detection. It also allows a parent menu class to take into
    account additional context e.g. shifting focus from one button to another using a keyboard or
    joystick input. Operations like this would be beyond the scope of a single Button instance.
    """

    is_focused: bool  # does the button have focus? (e.g. mouse hovering over)
    is_pressed: bool  # is the button down right now
    font_name = "ubuntu"
    font_size = 20
    text_color: Color

    def __init__(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        text=None,
        on_press=None,
        on_focus=None,
        on_release=None,
        on_unfocus=None,
    ):
        self.width = width
        self.height = height
        self.text = text
        self.rect = Rect(0, 0, width, height)
        self.rect.center = (x, y)
        self.is_focused = False
        self.is_pressed = False
        self._on_press = on_press or (lambda button: None)
        self._on_focus = on_focus or (lambda button: None)
        self._on_release = on_release or (lambda button: None)
        self._on_unfocus = on_unfocus or (lambda button: None)
        self.font = pygame.font.Font(pygame.font.get_default_font(), self.font_size)
        self.state = self.state_idle
        self.image = pygame.Surface(self.rect.size)
        self.image.fill(Color("red"))
        if self.text:
            text = self.font.render(self.text, True, Color("black"))
            textRect = text.get_rect()
            textRect.center = self.image.get_rect().center
            self.image.blit(text, textRect)
        super().__init__()

    # =============================================================================================
    # state functions handle behaviour that happens *every tick* the button is in that state
    # =============================================================================================
    def state_idle(self):
        if self.is_focused:
            self.focus()
        if self.is_pressed:
            self.press()

    def state_focus(self):
        if not self.is_focused:
            self.unfocus()
        if self.is_pressed:
            self.press()

    def state_press(self):
        if not self.is_pressed:
            self.release()

    # =============================================================================================
    # on_* functions handle behaviour that happens *once* at the transition to a new state.
    # By default they just call the hook passed in by the button creator, but Button
    # subclasses can also extend the method to add additional custom behaviour. External effects
    # (e.g. incrementing some value stored in a parent class) should be handled by the hook.
    # Internal effects (i.e. those that require access to the button instance) should be handled
    # by an overridden method on a subclass. A mix of both (e.g. adding particles to the parent
    # class at the button's position) could be solved using an Event.
    # =============================================================================================

    def on_press(self):
        self._on_press(self)

    def on_release(self):
        self._on_release(self)

    def on_focus(self):
        self._on_focus(self)

    def on_unfocus(self):
        self._on_unfocus(self)

    # =============================================================================================
    # courtesy methods to make switching states easy
    # =============================================================================================

    def press(self):
        self.on_press()
        self.state = self.state_press

    def focus(self):
        self.on_focus()
        self.state = self.state_focus

    def release(self):
        self.on_release()
        self.state = self.state_focus if self.is_focused else self.state_idle

    def unfocus(self):
        self.on_unfocus()
        self.state = self.state_idle


class ColoredButton(Button):
    idle_color = (100, 0, 100)  # Color("purple")
    focus_color = Color("orange")
    press_color = Color("red")
    text_color = Color("white")

    def state_idle(self):
        super().state_idle()
        self.color = (pulsing_value(self.tick, 80, 150, 0.03), 30, 75)

    def state_focus(self):
        super().state_focus()
        self.color = (
            pulsing_value(self.tick, 180, 255, 0.3),
            pulsing_value(self.tick, 100, 163, 0.3),
            0,
        )

    def state_press(self):
        super().state_press()
        self.color = self.press_color



from pygame.color import Color
import glob
from pathlib import Path

import pygame
from pygame import Surface

from robingame.utils import limit_value


def init_display() -> Surface:
    if not pygame.display.get_init():
        pygame.display.init()
        return pygame.display.set_mode((1, 1))
    else:
        return pygame.display.get_surface()


def load_image(filename, colorkey=None) -> Surface:
    init_display()
    try:
        image = pygame.image.load(filename)
    except pygame.error:
        print("Unable to load image:", filename)
        raise

    # colorkey needs to be set before .convert_alpha() is called, because Surfaces with a
    # per-pixel transparency (i.e. after convert_alpha) ignore colorkey.
    if colorkey is not None:
        if colorkey == -1:
            colorkey = image.get_at((0, 0))
        image.set_colorkey(colorkey, pygame.RLEACCEL)

    image = image.convert_alpha()
    return image


def not_empty(surface: Surface) -> bool:
    """Check if a surface has any non-zero pixels. `Surface.get_bounding_rect()` returns the
    smallest rectangle on the surface containing data. If the surface is empty, it will return
    Rect(0, 0, 0, 0), for which `any` returns False"""
    return any(surface.get_bounding_rect())


def empty_image(*args, **kwargs) -> Surface:
    img = Surface(*args, **kwargs).convert_alpha()
    img.fill((0, 0, 0, 0))
    return img


def relative_folder(current_file: str, folder: str) -> Path:
    return Path(current_file).parent.absolute() / folder


def pad_alpha(colour_tuple):
    if len(colour_tuple) == 3:
        # if no alpha channel supplied, assume it's full opacity
        return (*colour_tuple, 255)
    elif len(colour_tuple) == 4:
        return colour_tuple
    else:
        raise Exception("bogus colour, man")


def brighten_color(color: Color, amount: int) -> Color:
    color = Color(color)
    r = limit_value(color.r + amount, between=(0, 255))
    g = limit_value(color.g + amount, between=(0, 255))
    b = limit_value(color.b + amount, between=(0, 255))
    return Color(r, g, b, color.a)


def brighten(image: Surface, amount: int):
    width, height = image.get_size()
    # surface.copy() inherits surface's colorkey; preserving transparency
    new_image = image.copy()

    # iterate over all the pixels in the old surface, and write a pixel to the new surface in the
    # corresponding position. If the colour of the present pixel has an entry in the
    # color_mapping dict, then write the new colour instead of the old one.
    for x in range(width):
        for y in range(height):
            color = image.get_at((x, y))[:]
            new_color = brighten_color(color, amount)
            if new_color:
                new_image.set_at((x, y), pygame.Color(*new_color))
            else:
                new_image.set_at((x, y), pygame.Color(*color))

    return new_image


def scale_image(image: Surface, scale: float):
    x_scale = image.get_rect().width * scale
    y_scale = image.get_rect().height * scale
    image = pygame.transform.scale(image, (x_scale, y_scale))
    return image


def scale_images(images: [Surface], scale: float) -> [Surface]:
    return [scale_image(image, scale) for image in images]


def flip_image(image: Surface, flip_x=False, flip_y=False):
    return pygame.transform.flip(image, bool(flip_x), bool(flip_y))


def flip_images(images: [Surface], flip_x=False, flip_y=False):
    return [flip_image(image, flip_x, flip_y) for image in images]


def recolor_image(surface: Surface, color_mapping: dict) -> [Surface]:

    # make sure the colourmap has alpha channel on all colours
    color_mapping = {pad_alpha(k): pad_alpha(v) for k, v in color_mapping.items()}
    width, height = surface.get_size()
    # surface.copy() inherits surface's colorkey; preserving transparency
    new_surface = surface.copy()

    # iterate over all the pixels in the old surface, and write a pixel to the new surface in the
    # corresponding position. If the colour of the present pixel has an entry in the
    # color_mapping dict, then write the new colour instead of the old one.
    for x in range(width):
        for y in range(height):
            color = surface.get_at((x, y))[:]
            new_color = color_mapping.get(color)
            if new_color:
                new_surface.set_at((x, y), pygame.Color(*new_color))
            else:
                new_surface.set_at((x, y), pygame.Color(*color))

    return new_surface


def recolor_images(images: [Surface], colormap: dict) -> [Surface]:
    return [recolor_image(image, colormap) for image in images]


def load_spritesheet(
    filename: Path | str,
    image_size: (int, int) = None,
    colorkey=None,
    num_images: int = 0,
) -> [Surface]:
    """Load the image file. Don't call this until pygame.display has been initiated. Split
    the spritesheet into images and return a list of images.

    If image_size is None, load the whole spritesheet as one sprite.
    """
    filename = Path(filename)
    if not filename.exists():
        raise FileNotFoundError(f"Couldn't find {filename}")
    sheet = load_image(filename.as_posix(), colorkey)

    if image_size:
        width, height = image_size
        num_horizontal = sheet.get_rect().width // width
        num_vertical = sheet.get_rect().height // height
        rects = [
            pygame.Rect((width * i, height * j, width, height))
            for j in range(num_vertical)
            for i in range(num_horizontal)
        ]
        images = [sheet.subsurface(rect) for rect in rects]
        images = list(filter(not_empty, images))
        if num_images:
            images = images[:num_images]
    else:
        images = [sheet]
    return images


def load_image_sequence(filename: Path | str, colorkey=None, num_images: int = 0) -> [Surface]:
    """Load a sequence of images."""
    filename = Path(filename)
    parent_folder = filename.parent
    pattern = filename.stem
    files = glob.glob(f"{parent_folder}/{pattern}*")
    if not files:
        raise FileNotFoundError(f"Couldn't find {filename}")
    images = [load_image(file, colorkey) for file in files]
    if num_images:
        images = images[:num_images]
    return images



from pathlib import Path
from typing import Callable

from pygame import Surface

from robingame.image.utils import (
    load_spritesheet,
    load_image_sequence,
    flip_images,
    recolor_images,
    scale_images,
)


class SpriteAnimation:
    """
    Animates a sequence of images.
    Can scale, flip, and recolor itself.
    """

    images: list[Surface] | None
    source: Callable | None

    def __init__(
        self,
        images: [Surface] = None,
        scale: float = None,
        flip_x: bool = False,
        flip_y: bool = False,
        colormap: dict = None,
    ):
        self.images = images
        if scale:
            self.scale(scale)
        if flip_x or flip_y:
            self.flip(flip_x, flip_y)
        if colormap:
            self.recolor(colormap)

    @classmethod
    def from_image(
        cls,
        filename: Path | str,
        colorkey=None,
        scale: float = None,
        flip_x: bool = False,
        flip_y: bool = False,
        colormap: dict = None,
    ) -> "SpriteAnimation":
        return cls.from_spritesheet(
            filename=filename,
            image_size=None,
            colorkey=colorkey,
            flip_x=flip_x,
            flip_y=flip_y,
            colormap=colormap,
            scale=scale,
        )

    @classmethod
    def from_spritesheet(
        cls,
        filename: Path | str,
        image_size: (int, int),
        colorkey=None,
        num_images: int = 0,
        scale: float = None,
        flip_x: bool = False,
        flip_y: bool = False,
        colormap: dict = None,
    ) -> "SpriteAnimation":
        images = load_spritesheet(
            filename=filename, image_size=image_size, colorkey=colorkey, num_images=num_images
        )
        return cls(images=images, scale=scale, flip_x=flip_x, flip_y=flip_y, colormap=colormap)

    @classmethod
    def from_image_sequence(
        cls,
        filename: Path | str,
        colorkey=None,
        num_images: int = 0,
        scale: float = None,
        flip_x: bool = False,
        flip_y: bool = False,
        colormap: dict = None,
    ) -> "SpriteAnimation":
        images = load_image_sequence(filename=filename, colorkey=colorkey, num_images=num_images)
        return cls(images=images, scale=scale, flip_x=flip_x, flip_y=flip_y, colormap=colormap)

    ############## playback ###############
    def play(self, n: int):
        """
        Fetch frame with index n. This is used in the game loop (where n is the iteration
        counter) to animate the sprite. Return False when we've run out of frames.
        """
        try:
            return self.images[n]
        except IndexError:
            return False

    def loop(self, n: int):
        """If n is greater than the number of frames, start again at the beginning."""
        return self.play(n % len(self.images))

    def play_once(self, n: int, repeat_frame=-1):
        """Run the animation once and then continue returning the specified frame (default=last
        frame)."""
        try:
            return self.images[n]
        except IndexError:
            return self.images[repeat_frame]

    ############## edit in place ###############
    def flip(self, flip_x: bool, flip_y: bool):
        self.images = flip_images(self.images, flip_x, flip_y)

    def recolor(self, colormap: dict):
        self.images = recolor_images(self.images, colormap)

    def scale(self, scale: float):
        self.images = scale_images(self.images, scale)

    ############## edit and copy ###############
    def flipped_copy(self, flip_x=False, flip_y=False):
        return self.__class__(images=flip_images(self.images, flip_x, flip_y))

    def recolored_copy(self, colormap: dict):
        return self.__class__(images=recolor_images(self.images, colormap))

    def scaled_copy(self, scale: float):
        return self.__class__(images=scale_images(self.images, scale))

    def __len__(self):
        return len(self.images)


class SpriteDict(dict):
    """
    Takes a folder and list of files as input, and handles the creation of SpriteAnimations.
    """

    def __init__(
        self,
        folder: Path | str,
        size: (int, int) = None,
        file_mapping: dict = None,
        scale: int = 1,
        colormap: dict = None,
        create_flipped_versions: bool = False,
        type="spritesheet",
    ):
        """
        Create SpriteSheet for each file, but don't trigger .load() yet.
        """
        super().__init__()
        self.scale = scale
        self.image_size = size
        self.colormap = colormap
        self.create_flipped_versions = create_flipped_versions
        folder = Path(folder)
        self.folder = folder
        self.type = type
        self.sprite_sheets = dict()
        if file_mapping:
            for key, filename in file_mapping.items():
                self.register(**{key: filename})

    def register(self, **kwargs):
        for key, filename in kwargs.items():
            if self.type == "spritesheet":  # todo: add other options
                self[key] = unflipped = SpriteAnimation.from_spritesheet(
                    self.folder / filename,
                    image_size=self.image_size,
                    colormap=self.colormap,
                    scale=self.scale,
                )
                if self.create_flipped_versions:
                    self[f"{key}_right"] = unflipped
                    self[f"{key}_left"] = SpriteAnimation.from_spritesheet(
                        self.folder / filename,
                        image_size=self.image_size,
                        colormap=self.colormap,
                        scale=self.scale,
                        flip_x=True,
                    )



import pygame.mouse

from robingame.gui.button import Button
from robingame.objects import Game, Group, Particle
from robingame.utils import mouse_hovering_over, random_int


class ComplicatedOnPressHookExample(Game):
    """
    This is an example where the on_press hook adds particles to the outer Game class at the
    position of the Button instance. To do this it needs access to both the Game and Button
    instances.
    """

    screen_color = (50, 50, 50)

    def __init__(self):
        super().__init__()
        self.buttons = Group()
        self.particles = Group()
        self.child_groups += [self.buttons, self.particles]
        self.buttons.add(
            Button(
                x=200,
                y=100,
                width=400,
                height=50,
                text="press and hold for smoke",
                on_press=(
                    lambda button: (
                        self.particles.add(Flash(x=button.x, y=button.y)),
                        self.particles.add(Glow(x=button.x, y=button.y) for _ in range(10)),
                    )
                ),
                on_release=(lambda button: self.particles.kill()),
            )
        )

    def update(self):
        super().update()

        # button manager stuff
        for button in self.buttons:
            if mouse_hovering_over(button):
                button.is_focused = True
                button.is_pressed = pygame.mouse.get_pressed()[0]
            else:
                button.is_focused = False
                button.is_pressed = False


class Flash(Particle):
    gravity = 0.1
    friction = 0.1
    decay = 20
    color = (150, 200, 150)
    radius = 100


class Glow(Particle):
    gravity = 0.1
    friction = 0.1
    decay = 0.1
    color = (20, 20, 20)
    radius = 50

    def __init__(self, x, y):
        super().__init__(x, y, u=random_int(-5, 5), v=random_int(-5, 5))


if __name__ == "__main__":
    ComplicatedOnPressHookExample().main()



import pygame
from pygame import Color
from pygame import Surface
from pygame.rect import Rect
from pygame.sprite import AbstractGroup

from robingame.objects.group import Group


class Entity(pygame.sprite.Sprite):
    """
    Finite State Machine:
    - self.tick is incremented every time the main game loop executes
    - self.state is executed every tick
    - when self.state changes, self.tick is set to 0

    Hierarchical structure:
    - Entities can be added to Groups to create a hierarchical structure
    - The order of groups in the .groups attribute determines the draw order; it's basically the
    layers
    """

    _state: "method" = lambda *args, **kwargs: None
    child_groups: list  # groups of child Entities belonging to this entity
    parent_groups: list  # groups of which this Entity is a member
    tick: int = 0  # iterations of the main game loop
    parental_name = "parent"
    debug_color = Color("red")

    def __init__(self, *groups: AbstractGroup) -> None:
        super().__init__(*groups)
        self.child_groups = []

    def update(self):
        self.state()
        for group in self.child_groups:
            group.update()
        self.tick += 1

    def draw(self, surface: Surface, debug: bool = False):
        for group in self.child_groups:
            group.draw(surface, debug)

    def kill(self):
        """Removes self from all groups it is a member of."""
        for group in self.child_groups:
            group.kill()
        super().kill()

    def add_to_group(self, *objects, group: Group):
        """
        Add an object to one of self.child_groups and give the object a reference to self as parent.
        This method is intended to be used by more specific methods e.g.:
        def add_particle(*objects):
            self.add_to_group(*objects, self.particles)
        """
        group.add(*objects)
        # give the object a reference to this scene
        for obj in objects:
            setattr(obj, self.parental_name, self)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        """Reset self.tick when state changes so we know how long we've been in the current
        state."""
        self._state = new_state
        self.tick = 0

    @property
    def parent_groups(self):
        return super().groups()

    @property
    def groups(self):
        raise Exception("NO")

    def __repr__(self):
        # The `_Sprite__g` is necessary because of name mangling in subclasses I think
        return f"<{self.__class__.__name__} Entity(in {len(self._Sprite__g)} groups)>"


class PhysicalEntity(Entity):
    """
    attributes:
    - rect: used for collision detection and positioning
    - image: used for blitting to screen
    """

    level: Entity  # parent Entity
    image: Surface = None
    rect: Rect
    frame_duration: int  # higher = slower animation framerate

    def draw(self, surface: Surface, debug: bool = False):
        if self.image:
            surface.blit(self.image, self.image_rect)
        if debug:
            pygame.draw.rect(surface, self.debug_color, self.rect, 1)
            pygame.draw.circle(surface, self.debug_color, self.rect.center, 2, 1)
        super().draw(surface, debug)

    @property
    def image_rect(self):
        """Default is to align the image with the center of the object"""
        if self.image:
            image_rect = self.image.get_rect()
            image_rect.center = self.rect.center
            return image_rect
        else:
            return None

    @property
    def x(self):
        return self.rect.centerx

    @x.setter
    def x(self, new_value):
        new_value = round(new_value)
        self.rect.centerx = new_value

    @property
    def y(self):
        return self.rect.centery

    @y.setter
    def y(self, new_value):
        new_value = round(new_value)
        self.rect.centery = new_value

    @property
    def animation_frame(self):
        """Convert game ticks to animation frames."""
        return self.tick // self.frame_duration



from pathlib import Path

import pygame


def init_sound(
    frequency=44100,
    size=-16,
    channels=2,
    buffer=0,
):
    if not pygame.mixer.get_init():
        pygame.mixer.pre_init(frequency, size, channels, buffer)
        pygame.mixer.init()


def load_sound(filename: str | Path, volume=0.5, **sound_kwargs):
    filename = Path(filename)
    if not filename.exists():
        raise FileNotFoundError(filename.as_posix())

    init_sound(**sound_kwargs)
    sound = pygame.mixer.Sound(filename.as_posix())
    sound.set_volume(volume)
    return sound



import logging
import functools
import glob
import multiprocessing
import os
import subprocess
import sys
import time
from collections import deque
from pathlib import Path

import pygame
from pygame import Surface
from robingame.objects import Game

logger = logging.getLogger(__file__)


def clean_empty_recordings_dir(output_dir: Path):
    try:
        os.mkdir(output_dir.as_posix())
    except FileExistsError:
        pass

    # clear old files
    files = glob.glob((output_dir / "*").as_posix())
    for file in files:
        os.remove(file)


def save_image_async(filename, img_string: bytes, size: tuple[int, int], output_dir: Path):
    image = pygame.image.fromstring(img_string, size, "RGBA")
    pygame.image.save(image, str(output_dir / f"{filename}.png"))


def save_images_async(images: list[Surface], output_dir: Path):
    stringified = (
        (
            pygame.image.tostring(image, "RGBA"),
            image.get_size(),
        )
        for image in images
    )
    with multiprocessing.Pool(processes=8) as pool:
        result = pool.starmap(
            func=save_image_async,
            iterable=(
                (ii, string, size, output_dir) for ii, (string, size) in enumerate(stringified)
            ),
        )
    return result


def create_videos(output_dir: Path, filename: str):
    mp4_file = output_dir / f"{filename}.mp4"
    gif_file = output_dir / f"{filename}.gif"
    subprocess.run(
        [
            "ffmpeg",
            "-r",
            "60",
            "-i",
            str(output_dir / "%d.png"),
            "-r",
            "60",
            str(mp4_file),
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.STDOUT,
    )
    print(f"Created {mp4_file}")
    subprocess.run(
        [
            "ffmpeg",
            "-r",
            "60",
            "-i",
            str(output_dir / "%d.png"),
            "-filter_complex",
            # credit: https://superuser.com/questions/1049606/reduce-generated-gif-size-using-ffmpeg
            (
                "fps=30,"
                "scale=1080:-1:flags=lanczos,"
                "split[s0][s1];[s0]"
                "palettegen=max_colors=32[p];[s1][p]"
                "paletteuse=dither=bayer"
            ),
            str(gif_file),
        ],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.STDOUT,
    )
    print(f"Created {gif_file}")


def decorate_draw(func, screenshots):
    @functools.wraps(func)
    def wrapped(self, surface: Surface, debug: bool = False):
        # normal _draw call
        func(self, surface, debug)

        # also do _draw onto a new screenshot surface which we store
        screenshot = Surface(surface.get_size()).convert_alpha()
        func(self, screenshot, debug)
        screenshots.append(screenshot)

    return wrapped


def decorate_main(func, screenshots: deque, output_dir: Path, filename: str):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        """Run the game as normal, but intercept the quit signal and save all the screenshots to
        file."""
        try:
            func(*args, **kwargs)
        except SystemExit:
            print("Deleting old images/videos...")
            clean_empty_recordings_dir(output_dir)
            print(f"Saving {len(screenshots)} images...")
            t1 = time.perf_counter()
            save_images_async(screenshots, output_dir)
            t2 = time.perf_counter()
            print(f"Images saved in {t2-t1}s")
            print("Creating videos...")
            create_videos(output_dir, filename)
        pygame.quit()
        sys.exit()

    return wrapped


def record(
    cls: Game = None,
    *,
    n_frames: int,
    output_dir: Path,
    filename="out",
):
    """
    Patch the Game's ._draw() and .main() methods so that we keep a screenshot of every frame,
    which we later stitch into videos.
    """
    output_dir = output_dir or Path(__file__).parent / "recordings"
    screenshots = deque(maxlen=n_frames)

    def decorate(cls):
        cls._draw = decorate_draw(cls._draw, screenshots=screenshots)
        cls.main = decorate_main(
            cls.main,
            screenshots=screenshots,
            output_dir=output_dir,
            filename=filename,
        )
        return cls

    return decorate(cls) if cls else decorate



import time
from collections import deque

import pygame
from pygame.color import Color
from pygame.surface import Surface

from robingame.objects.entity import Entity
from robingame.text.font import fonts


class FpsTracker(Entity):
    """Displays a little number in the top right of the window if debug=True"""

    buffer_length = 60
    font = fonts.cellphone_black

    def __init__(self, *groups) -> None:
        super().__init__(*groups)
        self.queue = deque(maxlen=self.buffer_length)

    def update(self):
        t = time.perf_counter()
        self.queue.append(t)
        seconds = self.queue[-1] - self.queue[0]
        self.fps = int(self.buffer_length / seconds) if seconds else 0

    def draw(self, surface: Surface, debug: bool = False):
        if debug:
            scale = 2
            width = 50 * scale
            surf = Surface((width, width // 4))
            surf.fill(Color("white"))
            self.font.render(surf, f"FPS: {self.fps}", wrap=width, align=0, scale=scale)
            rect = surf.get_rect()
            rect.right = surface.get_rect().right - 10
            rect.top = surface.get_rect().top + 10
            surface.blit(surf, rect)



import pygame


class Group(pygame.sprite.Group):
    """Container for multiple sprite objects."""

    def update(self, *args):
        super().update(*args)

    def draw(self, surface, debug=False):
        """Draws all of the member sprites onto the given surface."""
        sprites = self.sprites()
        for sprite in sprites:
            sprite.draw(surface, debug)
        self.lostsprites = []

    def kill(self):
        """Kill all the sprites in this group. This is different from .empty().
        empty() does not kill the sprites in other groups."""
        for sprite in self:
            sprite.kill()



import string
from pathlib import Path

from robingame.image import init_display
from robingame.text.font import Font

window = init_display()

assets = Path(__file__).parent.parent / "assets"

test_font = Font(
    filename=assets / "test_font.png",
    image_size=(16, 16),
    letters=(
        string.ascii_uppercase
        + string.ascii_lowercase
        + "1234567890-=!@#$%^&*()_+[]\;',./{}|:\"<>?~`"
    ),
    trim=True,
    xpad=1,
    space_width=8,
)

cellphone_black = Font(
    filename=assets / "cellphone-black.png",
    image_size=(7, 9),
    letters=(
        """!"#$%&'()*+,-./0123456789:'<=>?@"""
        + string.ascii_uppercase
        + "[\]^_`"
        + string.ascii_lowercase
        + "{|}~"
    ),
    xpad=1,
    colorkey=-1,
    trim=True,
    space_width=4,
)
cellphone_white = Font(
    filename=assets / "cellphone-white.png",
    image_size=(7, 9),
    letters=(
        """!"#$%&'()*+,-./0123456789:'<=>?@"""
        + string.ascii_uppercase
        + "[\]^_`"
        + string.ascii_lowercase
        + "{|}~"
    ),
    xpad=1,
    colorkey=-1,
    trim=True,
    space_width=4,
)
cellphone_white_mono = Font(
    filename=assets / "cellphone-white.png",
    image_size=(7, 9),
    letters=(
        """!"#$%&'()*+,-./0123456789:'<=>?@"""
        + string.ascii_uppercase
        + "[\]^_`"
        + string.ascii_lowercase
        + "{|}~"
    ),
    colorkey=-1,
)



from robingame.image import SpriteAnimation, relative_folder

folder = relative_folder(__file__, "")


button_flash = SpriteAnimation.from_spritesheet(
    filename=folder / "flashybutton.png",
    image_size=(32, 16),
    scale=10,
)



import pygame.mouse

from robingame.animation import damping_response
from robingame.examples.gui_examples.assets import button_flash
from robingame.gui.button import Button
from robingame.image import scale_image, brighten, SpriteAnimation
from robingame.objects import Game, Group, Particle
from robingame.utils import mouse_hovering_over, random_int


class ButtonWithImages(Button):
    frame_duration = 3

    def __init__(self, x: int, y: int, width: int, height: int, **kwargs):
        super().__init__(x, y, width, height, **kwargs)
        self.image_idle = button_flash.images[-1]
        self.image_pressed = brighten(scale_image(self.image_idle.copy(), 0.9), amount=-50)
        self.image = self.image_idle

    def state_idle(self):
        self.image = self.image_idle
        super().state_idle()

    def state_focus(self):
        self.image = self.animation.play_once(self.animation_frame, repeat_frame=0)
        super().state_focus()

    def state_press(self):
        self.image = self.image_pressed
        super().state_press()

    def on_focus(self):
        super().on_focus()
        self.animation = button_flash

    def on_release(self):
        super().on_release()
        self.animation = SpriteAnimation(images=[self.image_idle])


class BouncyButton(Button):
    frame_duration = 3

    def __init__(self, x: int, y: int, width: int, height: int, **kwargs):
        super().__init__(x, y, width, height, **kwargs)
        self.image = button_flash.play(0)
        sizes = [damping_response(t) for t in range(20)]
        self.animation = button_flash
        self.animation_timer = 99
        self.physics_timer = 0
        self.amplitude = 0
        self.damping = 0.4

    def state_press(self):
        super().state_press()
        self.amplitude = -0.5
        self.physics_timer = 0  # holds the physics on "frame" 1
        # self.damping = 0.4

    def on_focus(self):
        super().on_focus()
        self.animation_timer = 0
        self.physics_timer = 0
        self.amplitude = 0.1
        # self.damping = 0.15

    def update(self):
        super().update()
        self.image = self.animation.play_once(self.animation_timer // self.frame_duration)
        scale_factor = damping_response(
            self.physics_timer, amp=self.amplitude, damping=self.damping, freq=0.2
        )
        if abs(scale_factor) > 0.005:  # prevents jittering
            self.image = scale_image(self.image.copy(), 1 + scale_factor)
        self.physics_timer += 1
        self.animation_timer += 1


class ColoredButtonEffectsExample(Game):
    """
    This is an example where the Buttons have different colours/animations for each state and
    transition.
    """

    screen_color = (50, 50, 50)
    window_width = 500
    window_height = 500

    def __init__(self):
        super().__init__()
        self.buttons = Group()
        self.particles = Group()
        self.child_groups += [self.buttons, self.particles]
        self.buttons.add(
            ButtonWithImages(
                x=200,
                y=100,
                width=200,
                height=100,
                text="press and hold for smoke",
                on_release=(
                    lambda button: (
                        self.particles.add(Flash(x=button.x, y=button.y)),
                        self.particles.add(Glow(x=button.x, y=button.y) for _ in range(50)),
                    )
                ),
                on_unfocus=(lambda button: self.particles.kill()),
            ),
            BouncyButton(x=200, y=300, width=200, height=100),
        )

    def update(self):
        super().update()

        # button manager stuff
        for button in self.buttons:
            if mouse_hovering_over(button):
                button.is_focused = True
                button.is_pressed = pygame.mouse.get_pressed()[0]
            else:
                button.is_focused = False
                button.is_pressed = False


class Flash(Particle):
    gravity = 0.1
    friction = 0.01
    decay = 20
    color = (150, 200, 150)
    radius = 100


class Glow(Particle):
    gravity = 0.1
    friction = 0.05
    decay = 0.1
    color = (20, 20, 20)
    radius = 50

    def __init__(self, x, y):
        super().__init__(x, y, u=random_int(-5, 5), v=random_int(-5, 5))


if __name__ == "__main__":
    ColoredButtonEffectsExample().main()



"""
# similar to pygame.key.get_pressed()
Usage:
```
controller1 = GamecubeController(1)
inputs = controller1.get_values()
if inputs[gamecube.A]:
    # do action for A
```
"""
import pygame

from robingame.input.queue import InputQueue

# button/input indices. These are used for lookup similarly to e.g. pygame.K_ESCAPE
# It doesn't really matter what order these are, so long as they match up to the order of the
# .get_values() method.
A = 0
B = 1
X = 2
Y = 3
Z = 4
L = 5
R = 6
START = 7
LEFT = 8
RIGHT = 9
UP = 10
DOWN = 11
C_LEFT = 12
C_RIGHT = 13
C_UP = 14
C_DOWN = 15
R_AXIS = 16
L_AXIS = 17
D_LEFT = 18
D_RIGHT = 19
D_UP = 20
D_DOWN = 21


def linear_map(input_value, input_range, output_range, limit_output=True):
    """Linearly map a set of inputs to a set of outputs. If limit_output==True, don't output any
    values outside the output range. Will still allow inputs outside the input range."""
    input_min, input_max = input_range
    output_min, output_max = output_range

    di = input_max - input_min
    do = output_max - output_min
    gradient = do / di
    offset = output_min - gradient * input_min

    # y = mx + c
    output_value = gradient * input_value + offset

    if limit_output:
        output_value = max(output_value, output_min)
        output_value = min(output_value, output_max)

    return output_value


class GamecubeControllerReader:
    """Class to read the inputs of a GameCube controller plugged into the Mayflash "GameCube
    Controller Adapter for Wii U & PC USB"

    This class
    - reads the values from the controller's buttons (binary values) axes (analog inputs such as
    control sticks and triggers) and hats (the 4-way D pad)
    - maps the values of the inputs to a 0-1 range and exposes these as named properties if the
    user wants to access a single value.
    - exposes a .get_values() method which returns a tuple of all the inputs (similar to how
    pygame deals with keyboard and mouse input.
    """

    # input ranges. Use these to set minimum (i.e. dead zone) and maximum input values
    GREY_STICK_INPUT_RANGE = (0.1, 0.86)
    YELLOW_STICK_INPUT_RANGE = (0.1, 0.86)
    TRIGGER_INPUT_RANGE = (-0.5, 1)

    def __init__(self, joystick_id: int):
        self.joystick = pygame.joystick.Joystick(joystick_id)  # get the joystick from pygame
        self.joystick.init()  # turn on the joystick

    def get_values(self):
        """Get the current state of all the inputs. This is intended to be equivalent to
        pygame.key.get_pressed so that the inputs can be processed in the same way."""
        # todo: the d-pad properties each access the d-pad values when they are called. So we're
        #  checking the status of the d-pad 4 times per tick when once would do. If the inputs
        #  ever get slow, maybe look at this.
        return (
            self.A,
            self.B,
            self.X,
            self.Y,
            self.Z,
            self.L,
            self.R,
            self.START,
            self.LEFT,
            self.RIGHT,
            self.UP,
            self.DOWN,
            self.C_LEFT,
            self.C_RIGHT,
            self.C_UP,
            self.C_DOWN,
            self.R_AXIS,
            self.L_AXIS,
            self.D_LEFT,
            self.D_RIGHT,
            self.D_UP,
            self.D_DOWN,
        )

    def grey_stick_map(self, input_value):
        return linear_map(input_value, self.GREY_STICK_INPUT_RANGE, (0, 1))

    def yellow_stick_map(self, input_value):
        return linear_map(input_value, self.YELLOW_STICK_INPUT_RANGE, (0, 1))

    def trigger_map(self, input_value):
        return linear_map(input_value, self.TRIGGER_INPUT_RANGE, (0, 1))

    # =================== AXES (=ANALOG INPUTS) =======================

    # uncalibrated internal values

    @property
    def _GREY_STICK_X_AXIS(self):
        return self.joystick.get_axis(0)

    @property
    def _GREY_STICK_Y_AXIS(self):
        return self.joystick.get_axis(1)

    @property
    def _YELLOW_STICK_X_AXIS(self):
        return self.joystick.get_axis(5)

    @property
    def _YELLOW_STICK_Y_AXIS(self):
        return self.joystick.get_axis(2)

    @property
    def _L_TRIGGER_AXIS(self):
        return self.joystick.get_axis(3)

    @property
    def _R_TRIGGER_AXIS(self):
        return self.joystick.get_axis(4)

    # calibrated external values

    @property
    def LEFT(self):
        return self.grey_stick_map(-self._GREY_STICK_X_AXIS)

    @property
    def RIGHT(self):
        return self.grey_stick_map(self._GREY_STICK_X_AXIS)

    @property
    def UP(self):
        return self.grey_stick_map(-self._GREY_STICK_Y_AXIS)

    @property
    def DOWN(self):
        return self.grey_stick_map(self._GREY_STICK_Y_AXIS)

    @property
    def C_LEFT(self):
        return self.yellow_stick_map(-self._YELLOW_STICK_X_AXIS)

    @property
    def C_RIGHT(self):
        return self.yellow_stick_map(self._YELLOW_STICK_X_AXIS)

    @property
    def C_UP(self):
        return self.yellow_stick_map(-self._YELLOW_STICK_Y_AXIS)

    @property
    def C_DOWN(self):
        return self.yellow_stick_map(self._YELLOW_STICK_Y_AXIS)

    @property
    def R_AXIS(self):
        return self.trigger_map(self._R_TRIGGER_AXIS)

    @property
    def L_AXIS(self):
        return self.trigger_map(self._L_TRIGGER_AXIS)

    # =================== 4-WAY SWITCHES =======================

    # uncalibrated internal values

    @property
    def _D_PAD(self):
        return self.joystick.get_hat(0)

    @property
    def _D_PAD_X(self):
        return self._D_PAD[0]

    @property
    def _D_PAD_Y(self):
        return self._D_PAD[1]

    # calibrated external values

    @property
    def D_LEFT(self):
        return int(self._D_PAD_X < 0)

    @property
    def D_RIGHT(self):
        return int(self._D_PAD_X > 0)

    @property
    def D_UP(self):
        return int(self._D_PAD_Y > 0)

    @property
    def D_DOWN(self):
        return int(self._D_PAD_Y < 0)

    # =================== BUTTONS (=BINARY VALUES) =======================

    # calibrated external values

    @property
    def X(self):
        return self.joystick.get_button(0)

    @property
    def A(self):
        return self.joystick.get_button(1)

    @property
    def B(self):
        return self.joystick.get_button(2)

    @property
    def Y(self):
        return self.joystick.get_button(3)

    @property
    def L(self):
        return self.joystick.get_button(4)

    @property
    def R(self):
        return self.joystick.get_button(5)

    @property
    def Z(self):
        return self.joystick.get_button(7)

    @property
    def START(self):
        return self.joystick.get_button(9)


class ButtonInput:
    def __init__(self, id: int, parent: InputQueue = None):
        """
        Class to describe a single input channel on a joystick/controller -- e.g. the "A" button
        on a gamecube controller. Implements methods which check with the parent input device
        whether this button is pressed, released, etc. This allows for the more pleasant shorthand:
        `controller.a_button.is_pressed` instead of `controller.is_pressed(controller.a_button)`

        :param int id: index of this input channel in the controller.get_values() tuple
        """
        self.id = id
        self.parent = parent

    @property
    def is_down(self):
        return self.parent.is_down(self.id)

    @property
    def is_pressed(self):
        return self.parent.is_pressed(self.id)

    @property
    def is_released(self):
        return self.parent.is_released(self.id)

    @property
    def value(self):
        """Does the same thing as is_down but makes some parts of the code more readable,
        especially for analog inputs that can be between 0-1."""
        return self.is_down

    def buffered_presses(self, buffer_length):
        return self.parent.buffered_presses(self.id, buffer_length)

    def buffered_releases(self, buffer_length):
        return self.parent.buffered_releases(self.id, buffer_length)

    def __sub__(self, other):
        return self.value - (other.value if isinstance(other, ButtonInput) else other)

    def __add__(self, other):
        return self.value + (other.value if isinstance(other, ButtonInput) else other)

    def __bool__(self):
        """This allows us to do `if input.UP` instead of `if input.UP.is_down`"""
        return bool(self.value)


class GamecubeController(InputQueue):
    """
    A wrapper around GamecubeControllerReader and InputQueue which leaves all the fiddly
    input-reading logic to GamecubeControllerReader, and provides a convenient interface for
    accessing the queue of inputs.

    This allows games to subclass this class and define new key mappings e.g. "A" --> "attack"
    """

    # input channels in CAPITALS to differentiate them from other methods
    LEFT = ButtonInput(LEFT)
    RIGHT = ButtonInput(RIGHT)
    UP = ButtonInput(UP)
    DOWN = ButtonInput(DOWN)
    A = ButtonInput(A)
    B = ButtonInput(B)
    X = ButtonInput(X)
    Y = ButtonInput(Y)
    C_UP = ButtonInput(C_UP)
    C_DOWN = ButtonInput(C_DOWN)
    C_LEFT = ButtonInput(C_LEFT)
    C_RIGHT = ButtonInput(C_RIGHT)
    START = ButtonInput(START)
    D_PAD_UP = ButtonInput(D_UP)
    L = ButtonInput(L_AXIS)
    R = ButtonInput(R_AXIS)

    def __init__(self, controller_id: int, queue_length=60):
        controller = GamecubeControllerReader(controller_id)
        self.controller_id = controller_id
        self.controller = controller
        super().__init__(queue_length)

        # for each parentless SingleInput declared on the class, create a new SingleInput
        # instance with self as parent.
        button_inputs = {
            name: attr
            for _class in self.__class__.__mro__
            for name, attr in _class.__dict__.items()
            if issubclass(_class, GamecubeController) and isinstance(attr, ButtonInput)
        }
        for name, attr in button_inputs.items():
            inp = ButtonInput(attr.id, parent=self)
            setattr(self, name, inp)

    def get_new_values(self):
        return self.controller.get_values()


if __name__ == "__main__":
    """visual input checking."""

    class TextPrint(object):
        def __init__(self):
            self.reset()
            pygame.font.init()
            self.font = pygame.font.Font(None, 30)

        def tprint(self, screen, textString):
            textBitmap = self.font.render(textString, True, BLACK)
            screen.blit(textBitmap, (self.x, self.y))
            self.y += self.line_height

        def reset(self):
            self.x = 10
            self.y = 10
            self.line_height = 20

        def indent(self):
            self.x += 10

        def unindent(self):
            self.x -= 10

    BLACK = pygame.Color("black")
    WHITE = pygame.Color("white")

    pygame.joystick.init()
    textPrint = TextPrint()

    screen = pygame.display.set_mode((500, 700))
    pygame.display.set_caption("My Game")
    clock = pygame.time.Clock()
    done = False
    while not done:
        for event in pygame.event.get():  # User did something.
            if event.type == pygame.QUIT:  # If user clicked close.
                done = True  # Flag that we are done so we exit this loop.

        screen.fill(WHITE)
        textPrint.reset()
        controller = GamecubeControllerReader(joystick_id=0)

        textPrint.indent()
        textPrint.tprint(screen, "Gamecube controller 0")

        textPrint.tprint(screen, "Buttons:")
        textPrint.indent()
        for button in "A B X Y Z L R START".split():
            textPrint.tprint(screen, f"{button}: {getattr(controller, button)}")
        textPrint.unindent()

        textPrint.tprint(screen, "Axes:")
        textPrint.indent()
        for button in (
            "_L_TRIGGER_AXIS _R_TRIGGER_AXIS "
            "_GREY_STICK_X_AXIS _GREY_STICK_Y_AXIS "
            "_YELLOW_STICK_X_AXIS "
            "_YELLOW_STICK_Y_AXIS "
        ).split():
            textPrint.tprint(screen, f"{button}: {getattr(controller, button)}")
        textPrint.unindent()

        textPrint.tprint(screen, "Hats:")
        textPrint.indent()
        for button in "_D_PAD".split():
            textPrint.tprint(screen, f"{button}: {getattr(controller, button)}")
        textPrint.unindent()

        textPrint.tprint(screen, "PROCESSED INPUTS:")
        textPrint.indent()
        for button in (
            "LEFT RIGHT UP DOWN "
            "C_LEFT C_RIGHT C_UP C_DOWN "
            "R_AXIS L_AXIS "
            "D_LEFT D_RIGHT D_UP D_DOWN "
        ).split():
            textPrint.tprint(screen, f"{button}: {getattr(controller, button)}")
        textPrint.unindent()

        # Go ahead and update the screen with what we've drawn.
        pygame.display.flip()

        # Limit to 20 frames per second.
        clock.tick(20)

    pygame.quit()



import sys

import pygame
from pygame.color import Color
from pygame.surface import Surface

from robingame.input import EventQueue
from robingame.objects.entity import Entity, Group
from robingame.objects.helpers import FpsTracker


class Game(Entity):
    """Special case of Entity; it is at the very top of the object tree."""

    fps: int = 60
    window_width: int = 500
    window_height: int = 500
    window_caption: str = ""
    font_name: str = "ubuntucondensed"
    font_size: int = 30
    parental_name = "game"
    screen_color = Color("black")
    debug = False

    def __init__(self):
        super().__init__()
        pygame.init()

        self.scenes = Group()
        self.fps_tracker = FpsTracker()
        self.child_groups = [self.scenes]

        pygame.font.init()
        self.font = pygame.font.Font(pygame.font.match_font(self.font_name), self.font_size)
        self.window = pygame.display.set_mode((self.window_width, self.window_height))
        pygame.display.set_caption(self.window_caption)
        self.clock = pygame.time.Clock()

    def main(self):
        """This is the outermost game function which runs once. It contains the outermost game
        loop. Here's where you should put your main event state machine."""
        self.running = True
        while self.running:
            self._update()
            self._draw(self.window, debug=self.debug)
        pygame.quit()
        sys.exit()

    def add_scene(self, *objects):
        self.add_to_group(*objects, group=self.scenes)

    def read_inputs(self):
        """I've put this in a separate method because I don't like the idea of putting the inputs
        in the same list as other child groups. The order might get ruined, or a subclass might
        overwrite the list. It's crucial that the inputs are read before updating."""
        EventQueue.update()
        for event in EventQueue.events:
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            if event.type == pygame.KEYDOWN and event.key == pygame.K_F1:
                self.debug = not self.debug

    def print_debug_info(self):
        pass

    def _update(self):
        """
        1. read inputs
        2. update
        """
        self.read_inputs()
        if self.debug:
            self.print_debug_info()
        self.update()
        self.fps_tracker.update()
        if self.fps:
            self.clock.tick(self.fps)

    def _draw(self, surface: Surface, debug: bool = False):
        surface.fill(self.screen_color)  # clear the screen
        self.draw(surface, debug)
        self.fps_tracker.draw(surface, debug)
        pygame.display.update()  # print to screen



import numpy
import pygame

from robingame.animation import ease_in_out
from robingame.objects import Entity, Group, EventQueue
from robingame.utils import mouse_hovering_over


class Menu(Entity):
    """Base menu class."""

    def __init__(self, *groups):
        super().__init__(*groups)
        self.buttons = Group()
        self.child_groups = [self.buttons]

    def add_button(self, *objects):
        self.add_to_group(*objects, group=self.buttons)

    def update(self):
        self.update_buttons()
        super().update()

    def state_idle(self):
        pass

    def update_buttons(self):
        """todo: if you wanted to make this really efficient, you could only perform updates if
        an event is detected."""
        mouse_click = any(event.type == pygame.MOUSEBUTTONDOWN for event in EventQueue.events)
        for button in self.buttons:
            button.is_focused = False
            button.is_pressed = False
            if mouse_hovering_over(button):
                if mouse_click:
                    button.is_pressed = True
                else:
                    button.is_focused = True


class MyMenu(Menu):
    button_width = 200
    button_height = 50
    button_size = (button_width, button_height)
    transition_length = 40

    def __init__(self, *groups):
        super().__init__(*groups)
        self.state = self.animate_in

    @property
    def game_rect(self):
        return self.game.window.get_rect()

    def arrange_buttons_vertically(self):
        num_buttons = len(self.buttons)
        top = self.game.window.get_rect().top + 150
        bottom = top + self.button_height * 1.5 * num_buttons
        ys = numpy.linspace(top, bottom, num=num_buttons)
        for button, y in zip(self.buttons, ys):
            button.y = y

    def animate_in(self):
        if self.tick == 0:
            self.arrange_buttons_vertically()
        try:
            # todo: maybe make these generators so that you're not generating the entire array
            #  each tick
            centerx = self.game_rect.centerx
            far_left = self.game_rect.left - self.button_width
            x = ease_in_out(x=self.tick, start=far_left, stop=centerx, num=self.transition_length)
            for button in self.buttons:
                button.x = x
        except IndexError:
            self.state = self.state_idle

    def animate_out(self, next_scene=None):
        def _animate_out():
            if next_scene:
                self.game.add_scene(next_scene)
            try:

                centerx = self.game_rect.centerx
                far_right = self.game_rect.right + self.button_width
                x = ease_in_out(
                    x=self.tick, start=centerx, stop=far_right, num=self.transition_length
                )
                for button in self.buttons:
                    button.x = x
            except IndexError:
                self.kill()

        return _animate_out



from dataclasses import dataclass, is_dataclass, asdict
from typing import Union

import pygame
from pygame.event import EventType, Event

from robingame.image import init_display

init_display()


class EventQueue:
    """Pygame's pygame.event.get() gets the events in the queue, but also empties the queue. This
    class solves that"""

    # this is intentional. I want to store the events on the class. Only one game will be active
    # at once, so we'll never need more than one instance of this class.
    events = []

    @classmethod
    def add(cls, event: Union[EventType, "dataclass"]):
        """
        Add the event to pygame's event queue, where it will stay until the .update() method
        is called to load it into cls.events.

        This prevents race conditions / order dependency where an event is added to the event
        queue and processed in the same tick.
        """
        if is_dataclass(event):
            event = Event(event.type, **asdict(event))
        pygame.event.post(event)

    @classmethod
    def update(cls):
        """
        Read all the events from pygame's event queue into cls.events
        (also clears pygame's event queue)
        """
        cls.events = pygame.event.get()

    @classmethod
    def filter(cls, **kwargs):
        return [
            event
            for event in cls.events
            if all(getattr(event, attribute, None) == value for attribute, value in kwargs.items())
        ]

    @classmethod
    def get(cls, **kwargs):
        try:
            return cls.filter(**kwargs)[0]
        except IndexError:
            return None



import pygame
from pygame import Rect
from pygame.color import Color

from robingame.objects.entity import PhysicalEntity
from robingame.utils import circle_surf


class Particle(PhysicalEntity):
    blit_flag = pygame.BLEND_RGB_ADD
    gravity: float
    friction: float
    decay: float
    color: Color
    radius: float

    def __init__(
        self,
        x,
        y,
        u=0,
        v=0,
        radius=None,
        color=None,
        gravity=None,
        friction=None,
        decay=None,
        blit_flag=None,
    ):
        super().__init__()
        self.color = self.color if color is None else color
        self.gravity = self.gravity if gravity is None else gravity
        self.friction = self.friction if friction is None else friction
        self.decay = self.decay if decay is None else decay
        self.blit_flag = self.blit_flag if blit_flag is None else blit_flag
        self.radius = self.radius if radius is None else radius
        self.rect = Rect(0, 0, 0, 0)
        self.x = x
        self.y = y
        self.u = u
        self.v = v

    def update(self):
        self.x += self.u
        self.y += self.v
        self.v += self.gravity
        self.u *= 1 - self.friction
        self.v *= 1 - self.friction
        self.radius -= self.decay
        if self.death_condition:
            self.kill()

    def draw(self, surface, debug=False):
        surf = circle_surf(round(self.radius), self.color)
        image_rect = surf.get_rect()
        image_rect.center = self.rect.center
        surface.blit(surf, image_rect, special_flags=self.blit_flag)

    @property
    def death_condition(self):
        return self.radius <= 0



from collections import deque

from robingame.utils import count_edges


class Empty(tuple):
    """Mock tuple of 0/1s that always returns a 0 no matter the index. This is used to
    spoof an empty pygame.key.get_pressed() tuple."""

    def __getitem__(self, *args, **kwargs) -> int:
        return 0


class InputQueue(deque):
    """
    Provides additional functionality beyond pygame.key.get_pressed().
    - Maintains a buffer of the last few inputs
    - Calculates which keys have been pressed and released this tick
    """

    def __init__(self, queue_length=5):
        super().__init__(maxlen=queue_length)

    def get_new_values(self) -> tuple[int]:
        """Subclasses should implement this. It should be something like
        pygame.key.get_pressed()"""
        raise NotImplementedError

    def read_new_inputs(self):
        self.append(self.get_new_values())

    def get_down(self) -> tuple[int]:
        """Return the keys which are currently held down"""
        return self[-1] if len(self) > 0 else Empty()

    def get_pressed(self) -> tuple[int]:
        """Return the keys that have just been pressed---i.e. those that are down this tick but
        not the previous tick"""
        try:
            current = self[-1]
            previous = self[-2]
            # we HAVE to use the __getattr__ method of the ScancodeWrapper
            # here. Using for/in to iterate over it gives erroneous results!
            # That's why I'm using the index to get the values.
            return tuple(
                int(current[i] and not previous[i])
                for (i, c), p in zip(enumerate(current), previous)
            )
        except IndexError:
            return Empty()

    def get_released(self) -> tuple[int]:
        """Return the keys that have just been released---i.e. those that are not down this
        tick, but were down the previous tick"""
        try:
            current = self[-1]
            previous = self[-2]
            return tuple(
                int(previous[i] and not current[i])
                for (i, c), p in zip(enumerate(current), previous)
            )
        except IndexError:
            return Empty()

    def is_pressed(self, key) -> int:
        """Check if a key has been pressed this tick"""
        keys = self.get_pressed()
        return keys[key]

    def is_down(self, key) -> int:
        """Check if a key is currently held down"""
        keys = self.get_down()
        return keys[key]

    def is_released(self, key) -> int:
        """Check if a key has been released this tick"""
        keys = self.get_released()
        return keys[key]

    def buffered_inputs(self, key, buffer_length):
        """Count the rising and falling edges. Can be used to detect past inputs."""
        buffer = list(self)[-buffer_length:]
        values = [layer[key] for layer in buffer]
        return count_edges(values)

    def buffered_presses(self, key, buffer_length):
        rising, falling = self.buffered_inputs(key, buffer_length)
        return rising

    def buffered_releases(self, key, buffer_length):
        rising, falling = self.buffered_inputs(key, buffer_length)
        return falling



import pygame.draw
from pygame.color import Color
from pygame.surface import Surface

from robingame.objects import Game
from robingame.text.font import fonts

snippet = """
 1space
  2space
Ook    in 
de     spelling 
van    sommige andere talen wordt de umlaut gebruikt, bijvoorbeeld in verwante 
talen als het IJslands en het Zweeds, om een soortgelijk 

klankverschijnsel
weer
te

geven, of in niet-verwante talen als het Fins, het Hongaars of het Turks, om dezelfde e-, eu- en u-klank weer te geven, die echter niet het resultaat van hetzelfde klankverschijnsel zijn.
"""


class FontTest(Game):
    window_width = 1500
    screen_color = (150, 150, 150)

    def draw(self, surface: Surface, debug: bool = False):
        super().draw(surface, debug)
        X = 500
        Y = 30
        WRAP = 500
        fonts.cellphone_white.render(
            surface,
            snippet,
            scale=2,
            wrap=WRAP,
            x=X,
            y=Y,
            align=-1,
        )
        pygame.draw.rect(surface, color=Color("red"), rect=(X, Y, WRAP, WRAP), width=1)


if __name__ == "__main__":
    FontTest().main()



import pygame

from robingame.input.queue import InputQueue


class KeyboardInputQueue(InputQueue):
    def get_new_values(self) -> tuple[int]:
        scancode_wrapper = pygame.key.get_pressed()
        return tuple(scancode_wrapper[ii] for ii in range(len(scancode_wrapper)))



import numpy
from typing import Callable


def ease(
    x: int,
    start: int | float,
    stop: int | float,
    num: int,
    function: Callable,
) -> float:
    """see https://easings.net/ for functions"""
    if x > num:
        raise ValueError(f"{x=} is greater than {num=}")
    distance = stop - start
    progress = x / num
    value = start + function(progress) * distance
    return value


def ease_in(x, start, stop, num, power=3) -> numpy.array:
    f = lambda x: x**power
    return ease(x=x, start=start, stop=stop, num=num, function=f)


def ease_out(x, start, stop, num, power=3) -> numpy.array:
    f = lambda x: 1 - (1 - x) ** power
    return ease(x=x, start=start, stop=stop, num=num, function=f)


def ease_in_out(x, start, stop, num, power=3) -> numpy.array:
    f = lambda x: 4 * x**power if x < 0.5 else 1 - (-2 * x + 2) ** power / 2
    return ease(x=x, start=start, stop=stop, num=num, function=f)


def damping_response(t, amp=0.5, damping=0.4, phase=0, freq=0.3):
    decay = damping * freq
    return amp * numpy.e ** (-decay * t) * numpy.cos(freq * t - phase)
